#include "DXUT.h"
#include "Physics/PhysicsApp.h"
#include "Script/LuaTable.h"
#include "Physics/PhysicsMathConversions.h"
#include "Physics/PhysicsEntity.h"
#include "Physics/PxEnumReflections.h"

using namespace Egg;
using namespace physx;

Physics::PhysicsApp::PhysicsApp(ID3D11Device* device)
	:ScriptedApp(device),simulationEventCallback(this)
{
	using namespace luabind;

	module(luaState)
	[
		class_<Physics::PhysicsEntity>("PhysicsEntity"),

		class_<Physics::PhysicsApp, Script::ScriptedApp>("PhysicsApp")
			.def("PhysicsMaterial", &Physics::PhysicsApp::addPhysicsMaterial)
			.def("PhysicsEntity", &Physics::PhysicsApp::addPhysicsEntity)
			.def("Shape", &Physics::PhysicsApp::addShapeToPhysicsEntity)
			.def("Dynamics", &Physics::PhysicsApp::setDynamicsForPhysicsEntity)

			.def("addForceAndTorque", &Physics::PhysicsApp::addForceToPhysicsEntity)
			.def("setForceAndTorqueForTarget", &Physics::PhysicsApp::setForceAndTorqueForTargetToPhysicsEntity)
			.def("spawn", &Physics::PhysicsApp::spawnPhysicsEntity)
	];

	call_function<Physics::PhysicsApp*>(luaState, "setEgg", this);

	PxEnumReflections::initialize();

	timestep = 0.02;
	timeRemainingOfTimestep = 0.0;
}

PxFilterFlags allContactsFilterShader(
        PxFilterObjectAttributes attributes0, PxFilterData filterData0,
        PxFilterObjectAttributes attributes1, PxFilterData filterData1,
        PxPairFlags& pairFlags, const void* constantBlock, PxU32 constantBlockSize)
{
        // generate contacts for all that were not filtered above
        pairFlags = PxPairFlag::eCONTACT_DEFAULT;
		pairFlags |= PxPairFlag::eNOTIFY_TOUCH_FOUND;

        return PxFilterFlag::eDEFAULT;
}


HRESULT Physics::PhysicsApp::createResources()
{
	__super::createResources();

	static PxDefaultErrorCallback gDefaultErrorCallback;
	static PxDefaultAllocator gDefaultAllocatorCallback;
	static PxSimulationFilterShader gDefaultFilterShader = PxDefaultSimulationFilterShader;
	foundation = PxCreateFoundation(PX_PHYSICS_VERSION, gDefaultAllocatorCallback, gDefaultErrorCallback);

	physics = PxCreatePhysics(PX_PHYSICS_VERSION, *foundation, PxTolerancesScale(), false);

	// check if PvdConnection manager is available on this platform
	if(physics->getPvdConnectionManager() != NULL)
	{
		// setup connection parameters
		const char*     pvd_host_ip = "127.0.0.1";  // IP of the PC which is running PVD
		int             port        = 5425;         // TCP port to connect to, where PVD is listening
		unsigned int    timeout     = 100;          // timeout in milliseconds to wait for PVD to respond,
													// consoles and remote PCs need a higher timeout.
		PxVisualDebuggerConnectionFlags connectionFlags = PxVisualDebuggerExt::getAllConnectionFlags();

		// and now try to connect
		pvdConnection = PxVisualDebuggerExt::createConnection(physics->getPvdConnectionManager(),
			pvd_host_ip, port, timeout, connectionFlags);

	}
	else
		pvdConnection = NULL;

	PxSceneDesc sceneDesc(physics->getTolerancesScale());
	sceneDesc.gravity = PxVec3(0.0f, -9.81f, 0.0f);
	if(!sceneDesc.cpuDispatcher)
	{
		cpuDispatcher = PxDefaultCpuDispatcherCreate(1);
		sceneDesc.cpuDispatcher    = cpuDispatcher;
	}
	if(!sceneDesc.filterShader)
		sceneDesc.filterShader    = allContactsFilterShader;//gDefaultFilterShader;


	scene = physics->createScene(sceneDesc);
	scene->setSimulationEventCallback(&simulationEventCallback);

	scene->simulate(0.01);
	
	return S_OK;
}

HRESULT Physics::PhysicsApp::releaseResources()
{
	if (pvdConnection)
		pvdConnection->release();

	std::vector<PhysicsEntity::W>::iterator iPhysicsEntity = physicsEntities.begin();
	while(iPhysicsEntity != physicsEntities.end())
	{
		PhysicsEntity::P physicsEntity = (*iPhysicsEntity).lock();
		if(physicsEntity)
			physicsEntity->releaseControlState();
		iPhysicsEntity++;
	}
	physicsEntities.clear();
	return __super::releaseResources();
}

void Physics::PhysicsApp::render(ID3D11DeviceContext* context)
{	
	__super::render(context);
}

void Physics::PhysicsApp::animate(double dt, double t)
{
	using namespace luabind;
	try
	{
		int iKey=0;
		for (iterator i(globals(luaState)["keysPressed"]), end; i != end; ++i, ++iKey)
		{
			std::string kks = luabind::object_cast<std::string>(i.key());
			int keycode = kks.at(0);
			if( kks == "VK_SPACE" )   keycode = VK_SPACE;
			if( kks == "VK_NUMPAD0" ) keycode = VK_NUMPAD0;
			if( kks == "VK_NUMPAD1" ) keycode = VK_NUMPAD1;
			if( kks == "VK_NUMPAD2" ) keycode = VK_NUMPAD2;
			if( kks == "VK_NUMPAD3" ) keycode = VK_NUMPAD3;
			if( kks == "VK_NUMPAD4" ) keycode = VK_NUMPAD4;
			if( kks == "VK_NUMPAD5" ) keycode = VK_NUMPAD5;
			if( kks == "VK_NUMPAD6" ) keycode = VK_NUMPAD6;
			if( kks == "VK_NUMPAD7" ) keycode = VK_NUMPAD7;
			if( kks == "VK_NUMPAD8" ) keycode = VK_NUMPAD8;
			if( kks == "VK_NUMPAD9" ) keycode = VK_NUMPAD9;

			*i = controlParameters.keyPressed[keycode]?true:false;
		}

		timeRemainingOfTimestep -= dt;
		if(timeRemainingOfTimestep < 0)
		{
			timeRemainingOfTimestep += timestep;
			globals(luaState)["dt"] = timestep;
			__super::animate(dt, t);
			entities.insert(spawnedEntities.begin(), spawnedEntities.end());
			spawnedEntities.clear();
			scene->simulate(timestep);
			scene->fetchResults(true);
		}
	}
	catch(luabind::error exception){ 
		std::string errs = lua_tostring(luaState, 0);
		MessageBoxA( NULL, errs.c_str(), "Script error!", MB_OK);
		exit(-1); }
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

bool Physics::PhysicsApp::processMessage( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	return __super::processMessage(hWnd, uMsg, wParam, lParam);
}

void Physics::PhysicsApp::addPhysicsMaterial(luabind::object nil, luabind::object attributes)
{
	Script::LuaTable attributeTable(attributes, "PhysicsMaterial");
	try
	{
		using namespace Egg::Math;
		std::string name = attributeTable.getString("name");
		PxMaterial* material = physics->createMaterial(
			attributeTable.getFloat("staticFriction", 0.5),
			attributeTable.getFloat("dynamicFriction", 0.5),
			attributeTable.getFloat("restitution", 0.5)
			);

		physicsMaterials[name] = material;
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

void Physics::PhysicsApp::addPhysicsEntity(luabind::object nil, luabind::object attributes, luabind::object initializer)
{
	using namespace Egg::Script;
	LuaTable attributeTable(attributes, "PhysicsEntity");
	try
	{
		std::string entityName = attributeTable.getString("name");
		std::string multiMeshName = attributeTable.getString("multiMesh");
		using namespace Egg::Math;
		float3 position = attributeTable.getFloat3("position");
		float3 axis = attributeTable.getFloat3("orientationAxis", float3(0, 1, 0));
		float angle = attributeTable.getFloat("orientationAngle");
		Scene::Directory<Mesh::Multi>::iterator iMultiMesh = multiMeshes.find(multiMeshName);
		if(iMultiMesh == multiMeshes.end())
			attributeTable.throwError(std::string("Unknown MultiMesh '") +multiMeshName+ "'.");
		Physics::PhysicsEntity::P physicsEntity = Physics::PhysicsEntity::create(scene, position, float4::quatAxisAngle(axis, angle), iMultiMesh->second);
		physicsEntity->setControlState(attributeTable.getLuaBindObject("controlState"));
		densityForEntityBeingAdded = 1.0;
		initializer(physicsEntity);
		if(PxRigidDynamic* rigidDynamic = physicsEntity->getActor()->isRigidDynamic())
			PxRigidBodyExt::updateMassAndInertia(*rigidDynamic, densityForEntityBeingAdded);
		physicsEntity->getActor()->userData = physicsEntity.get();
		scene->addActor(*physicsEntity->getActor());
		entities[entityName] = physicsEntity;
		physicsEntities.push_back(physicsEntity);
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

void Physics::PhysicsApp::addShapeToPhysicsEntity(PhysicsEntity::P physicsEntity, luabind::object attributes)
{
	using namespace Egg::Script;
	LuaTable attributeTable(attributes, "Shape");
	try
	{
		using namespace Egg::Math;
		using namespace physx;
		float3 position = attributeTable.getFloat3("position");
		float3 axis = attributeTable.getFloat3("orientationAxis", float3(0, 1, 0));
		axis = axis.normalize();
		float angle = attributeTable.getFloat("orientationAngle");

		std::string materialName = attributeTable.getString("material");
		PhysicsMaterialDirectory::iterator iMaterial = physicsMaterials.find(materialName);
		if(iMaterial == physicsMaterials.end())
			attributeTable.throwError(std::string("Unknown physics material '") +materialName+ "'.");
		PxGeometry* geometry = NULL;
		PxGeometryType::Enum e = attributeTable.getEnum<PxGeometryType::Enum>("geometryType", PxGeometryType::eINVALID);
		switch(e)
		{
			case PxGeometryType::eBOX:
				geometry = new PxBoxGeometry( ~attributeTable.getFloat3("halfExtents", float3(1, 1, 1)) );
				break;
			case PxGeometryType::eCAPSULE:
				geometry = new PxCapsuleGeometry( attributeTable.getFloat("radius"), attributeTable.getFloat("halfHeight") );
				break;
			case PxGeometryType::ePLANE:
				geometry = new PxPlaneGeometry( );
				break;
			case PxGeometryType::eSPHERE:
				geometry = new PxSphereGeometry( attributeTable.getFloat("radius") );
				break;
			case PxGeometryType::eINVALID:
				attributeTable.throwError("Geometry type not specified!");
			default:
				attributeTable.throwError("Geometry type not supported!");
		}
		physicsEntity->getActor()->createShape(
			*geometry, *iMaterial->second, PxTransform(~position, PxQuat(angle, ~axis)));
		delete geometry;
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

void Physics::PhysicsApp::setDynamicsForPhysicsEntity(PhysicsEntity::P physicsEntity, luabind::object attributes)
{
	using namespace Egg::Script;
	LuaTable attributeTable(attributes, "Dynamics");
	try
	{
		PxRigidDynamic* rigidDynamic = physicsEntity->makeActorDynamic(scene);
		rigidDynamic->setActorFlags(attributeTable.getEnumCombination<PxActorFlag::Enum, PxActorFlags>("actorFlags"));

		rigidDynamic->setLinearDamping( attributeTable.getFloat("linearDamping", rigidDynamic->getLinearDamping() ));
		rigidDynamic->setAngularDamping( attributeTable.getFloat("angularDamping", rigidDynamic->getAngularDamping() ));
		rigidDynamic->setMaxAngularVelocity( attributeTable.getFloat("maxAngularVelocity", rigidDynamic->getMaxAngularVelocity() ));
		rigidDynamic->setRigidDynamicFlags( attributeTable.getEnumCombination<PxRigidDynamicFlag::Enum, PxRigidDynamicFlags>("rigidDynamicFlags", rigidDynamic->getRigidDynamicFlags() ));
		rigidDynamic->setLinearVelocity( ~attributeTable.getFloat3("linearVelocity", ~rigidDynamic->getLinearVelocity() ));
		rigidDynamic->setAngularVelocity( ~attributeTable.getFloat3("angularVelocity", ~rigidDynamic->getAngularVelocity() ));
		float density = attributeTable.getFloat("density", -1.0);
		if(density > 0)
			densityForEntityBeingAdded = density;
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

void Physics::PhysicsApp::addForceToPhysicsEntity(PhysicsEntity::P physicsEntity, luabind::object attributes)
{
	try
	{
		using namespace Egg::Script;
		LuaTable attributeTable(attributes, "addForceAndTorque");
		PxRigidDynamic* rigidDynamic = physicsEntity->getActor()->isRigidDynamic();
		if(rigidDynamic)
		{
			rigidDynamic->addForce( ~((attributeTable.getFloat3("force").xyz1 * physicsEntity->getRotationMatrix()).xyz));
			rigidDynamic->addTorque(~((attributeTable.getFloat3("torque").xyz1 * physicsEntity->getRotationMatrix()).xyz));
		}
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

bool Physics::PhysicsApp::setForceAndTorqueForTargetToPhysicsEntity(PhysicsEntity::P physicsEntity, luabind::object attributes)
{
	try
	{
		using namespace Egg::Scene;
		using namespace Egg::Script;
		using namespace luabind;
		using namespace Egg::Math;

		PxRigidDynamic* rigidDynamic = physicsEntity->getActor()->isRigidDynamic();
		if(!rigidDynamic)
			return false;

		LuaTable attributeTable(attributes, "addForceAndTorqueToTarget");
		float proximityRadius = attributeTable.getFloat("proximityRadius");
		float maxForce = attributeTable.getFloat("maxForce");
		float maxTorque = attributeTable.getFloat("maxTorque");
		float3 markPosition = attributeTable.getFloat3("position");
		std::string markName = attributeTable.getString("mark", "");
		if(!markName.empty())
		{
			Directory<Entity>::iterator iEntity = entities.find(markName);
			//TODO: handle cases where mark has been destroyed
			if(iEntity == entities.end())
				attributeTable.throwError(std::string("Unknown entity ") + markName + " specified as mark.");
			markPosition = (markPosition.xyz1 * iEntity->second->getModelMatrix()).xyz;
		}
		float3 markDiff = markPosition - physicsEntity->getPosition();
		float markDist = markDiff.length();
		float3 markDir = markDiff / markDist;
		float3 ahead = (float3::xUnit.xyz0 * physicsEntity->getRotationMatrix()).xyz;
		rigidDynamic->addForce(~(ahead * markDir.dot(ahead) * maxForce ) );
		rigidDynamic->addTorque(~(ahead.cross(markDir) * maxTorque ) );

		if(markDist < proximityRadius)	return true;
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }

	return false;
}

void Physics::PhysicsApp::spawnPhysicsEntity(PhysicsEntity::P parentPhysicsEntity, luabind::object attributes, luabind::object initializer)
{
	using namespace Egg::Script;
	LuaTable attributeTable(attributes, "spawn");
	try
	{
		std::string entityName = attributeTable.getString("name");
		std::string multiMeshName = attributeTable.getString("multiMesh");
		using namespace Egg::Math;
		float3 position = attributeTable.getFloat3("position");
		position = (position.xyz1 * parentPhysicsEntity->getModelMatrix()).xyz;
		float3 axis = attributeTable.getFloat3("orientationAxis", float3(0, 1, 0));
		float angle = attributeTable.getFloat("orientationAngle");
		float4 orientation = parentPhysicsEntity->getOrientation().quatMul( float4::quatAxisAngle(axis, angle) );

		Scene::Directory<Mesh::Multi>::iterator iMultiMesh = multiMeshes.find(multiMeshName);
		if(iMultiMesh == multiMeshes.end())
			attributeTable.throwError(std::string("Unknown MultiMesh '") +multiMeshName+ "'.");
		Physics::PhysicsEntity::P physicsEntity = Physics::PhysicsEntity::create(scene, position, orientation, iMultiMesh->second);
		physicsEntity->setControlState(attributeTable.getLuaBindObject("controlState"));
		densityForEntityBeingAdded = 1.0;
		initializer(physicsEntity);
		if(PxRigidDynamic* rigidDynamic = physicsEntity->getActor()->isRigidDynamic())
		{
			rigidDynamic->setLinearVelocity( ~((~rigidDynamic->getLinearVelocity()).xyz1 * parentPhysicsEntity->getRotationMatrix()).xyz );
			rigidDynamic->setAngularVelocity( ~((~rigidDynamic->getAngularVelocity()).xyz1 * parentPhysicsEntity->getRotationMatrix()).xyz );

			PxRigidBodyExt::updateMassAndInertia(*rigidDynamic, densityForEntityBeingAdded);
		}
		physicsEntity->getActor()->userData = physicsEntity.get();
		scene->addActor(*physicsEntity->getActor());
		spawnedEntities[entityName] = physicsEntity;
		physicsEntities.push_back(physicsEntity);
	}
	catch(Egg::HrException exception){ exitWithErrorMessage(exception); }
}

void  Physics::PhysicsApp::SimulationEventCallback::onContact (const PxContactPairHeader &pairHeader, const PxContactPair *pairs, PxU32 nbPairs)
{
	for(PxU32 i=0; i < nbPairs; i++)
	{
		const PxContactPair& cp = pairs[i];

		if(cp.events & PxPairFlag::eNOTIFY_TOUCH_FOUND)
		{
			PhysicsEntity::P physicsEntity0 = ((PhysicsEntity*)pairHeader.actors[0]->userData)->shared_from_this();
			PhysicsEntity::P physicsEntity1 = ((PhysicsEntity*)pairHeader.actors[1]->userData)->shared_from_this();

			physicsEntity0->onContact(physicsEntity1);
			physicsEntity1->onContact(physicsEntity0);
		}
	}
}
